import logging
import unittest
from unittest import mock

import prov
from bigchaindb_driver import exceptions as bdb_exceptions

from prov2bigchaindb.core import utils, exceptions
from prov2bigchaindb.tests.core import setup_test_files

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class UtilityTest(unittest.TestCase):
    def setUp(self):
        self.test_prov_files = setup_test_files()
        self.account_id = 'Utility_Test_Account'

    def tearDown(self):
        del self.account_id
        del self.test_prov_files

    def test_to_provdocument(self):
        doc = prov.model.ProvDocument()
        doc.add_namespace('ex', 'http://example.org/')
        doc.entity('ex:foo')
        doc.agent('ex:Bob')
        doc.wasAttributedTo('ex:foo', 'ex:Bob')
        # test json
        prov_document = utils.to_prov_document(content=doc.serialize(format='json'))
        self.assertEqual(doc, prov_document)
        with self.assertRaises(exceptions.ParseException):
            utils.to_prov_document(content="{fooo")
        # test xml
        prov_document = utils.to_prov_document(content=doc.serialize(format='xml'))
        self.assertEqual(doc, prov_document)
        with self.assertRaises(exceptions.ParseException):
            utils.to_prov_document(content="<?xml")
        # test bytes
        content = self.test_prov_files["simple"]
        doc = prov.model.ProvDocument.deserialize(content=content, format='json').flattened()
        prov_document = utils.to_prov_document(content=self.test_prov_files["simple"])
        self.assertEqual(doc, prov_document)
        # test provn (not implement yet by prov)
        with self.assertRaises(NotImplementedError):
            utils.to_prov_document(content=doc.serialize(format='provn'))
        # test invalid document
        with self.assertRaises(exceptions.ParseException):
            utils.to_prov_document(content="fooo")

    @mock.patch('prov2bigchaindb.core.utils.BigchainDB')
    def test_wait_until_valid(self, mock_bdb):
        mock_bdb.transactions.retrieve.return_value = {'id': '1', 'asset': {
            'data': {'prov': ''}}}
        mock_bdb.transactions.status.return_value = {'status': 'valid'}
        utils.wait_until_valid('1', mock_bdb)
        mock_bdb.transactions.status.assert_called_once_with('1')
        mock_bdb.transactions.status.side_effect = bdb_exceptions.NotFoundError()
        with self.assertRaises(exceptions.TransactionIdNotFound):
            utils.wait_until_valid('1', mock_bdb)

    @mock.patch('prov2bigchaindb.core.utils.BigchainDB')
    def test_is_valid_tx(self, mock_bdb):
        mock_bdb.transactions.status.return_value = {'status': 'valid'}
        ret = utils.is_valid_tx('1', mock_bdb)
        mock_bdb.transactions.status.assert_called_once_with('1')
        self.assertEqual(ret, True)
        mock_bdb.transactions.status.return_value = {'status': 'backlog'}
        ret = utils.is_valid_tx('1', mock_bdb)
        self.assertEqual(ret, False)
        mock_bdb.transactions.status.return_value = {'status': 'undecided'}
        ret = utils.is_valid_tx('1', mock_bdb)
        self.assertEqual(ret, False)
        mock_bdb.transactions.status.side_effect = bdb_exceptions.NotFoundError()
        with self.assertRaises(exceptions.TransactionIdNotFound):
            utils.is_valid_tx('1', mock_bdb)

    @mock.patch('prov2bigchaindb.core.utils.BigchainDB')
    @mock.patch('prov2bigchaindb.core.utils.requests.models.Response')
    @mock.patch('prov2bigchaindb.core.utils.requests')
    @mock.patch('prov2bigchaindb.core.utils.random')
    def test_is_block_to_tx_valid(self, mock_random, mock_requests, mock_response, mock_bdb):
        mock_random.choice.return_value = 'http://127.0.0.1:9984'
        mock_response.json.side_effect = [['1'], {'status': 'valid'},
                                          ['1'], {'status': 'undecided'},
                                          ['1'], {'status': 'invalid'},
                                          [],
                                          ['1']]
        mock_response.status_code = 200
        mock_requests.get.return_value = mock_response
        ret = utils.is_block_to_tx_valid('1', mock_bdb)
        self.assertEqual(ret, True)
        ret = utils.is_block_to_tx_valid('1', mock_bdb)
        self.assertEqual(ret, False)
        ret = utils.is_block_to_tx_valid('1', mock_bdb)
        self.assertEqual(ret, False)
        with self.assertRaises(exceptions.TransactionIdNotFound):
            utils.is_block_to_tx_valid('1', mock_bdb)
        mock_response.status_code = 400
        with self.assertRaises(exceptions.BlockIdNotFound):
            utils.is_block_to_tx_valid('1', mock_bdb)



import logging
from datetime import datetime

from bigchaindb_driver import BigchainDB
from bigchaindb_driver.crypto import generate_keypair
from prov.model import ProvDocument, ProvElement, ProvAgent

from prov2bigchaindb.core import utils, exceptions, local_stores

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class BaseAccount(object):
    """
    BigchainDB Base Account
    """

    def __init__(self, account_id: str, store: local_stores.SqliteStore):
        """
        Instantiate BaseAccount object

        :param account_id: Internal id of Account
        :type account_id: str
        :param store: Local database object
        :type store: local_stores.SqliteStore
        """
        assert account_id is not None
        assert store is not None
        self.store = store
        self.account_id = account_id
        self.tx_id = ''
        self.private_key, self.public_key = generate_keypair()
        try:
            self.account_id, self.public_key, self.private_key, self.tx_id = self.store.get_account(self.account_id)
            log.debug("Found account for %s with public_key %s", self.account_id, self.public_key)
        except exceptions.NoAccountFoundException:
            self.store.write_account(self.account_id, self.public_key, self.private_key)
            log.debug("New account for %s with public_key %s", self.account_id, self.public_key)

    def __str__(self):
        return "{} : {}".format(self.account_id, self.public_key)

    def _create_asset(self, bdb_connection: BigchainDB, asset: dict, metadata: dict = None) -> dict:
        """
        Create and transfer new CREATE transaction

        :param bdb_connection: Connection object for BigchainDB
        :type bdb_connection: BigchainDB
        :param asset: Dictonary with asset data
        :type asset: dict
        :param metadata: Dictonary with additional metadata
        :type metadata: dict or None
        :return: Result with created CREATE transactions
        :rtype: dict
        """
        if metadata is None:
            metadata = {}
        metadata['timestamp'] = datetime.utcnow().timestamp()
        prepared_creation_tx = bdb_connection.transactions.prepare(operation='CREATE',
                                                                   signers=self.public_key,
                                                                   asset=asset,
                                                                   metadata=metadata)
        fulfilled_creation_tx = bdb_connection.transactions.fulfill(prepared_creation_tx,
                                                                    private_keys=self.private_key)
        sent_creation_tx = bdb_connection.transactions.send(fulfilled_creation_tx)
        if fulfilled_creation_tx != sent_creation_tx:
            raise exceptions.CreateRecordException()
        return sent_creation_tx

    def _transfer_asset(self, bdb_connection: BigchainDB, recipient_pub_key: str, tx: dict,
                        metadata: dict = None) -> dict:
        """
        Create and transfer new TRANSFER transaction

        :param bdb_connection: Connection object for BigchainDB
        :type bdb_connection: BigchainDB
        :param recipient_pub_key: Public key of the recipient
        :type recipient_pub_key: str
        :param tx: Transaction which should be transferd
        :type tx: dict
        :param metadata: Dictionary with additional metadata
        :type metadata: dict or None
        :return: Result with created CREATE transactions
        :rtype: dict
        """
        if metadata is None:
            metadata = {}
        metadata['timestamp'] = datetime.utcnow().timestamp()
        transfer_asset = {'id': tx['id']}
        output_index = 0
        output = tx['outputs'][output_index]
        transfer_input = {
            'fulfillment': output['condition']['details'],
            'fulfills': {
                'output': output_index,
                'transaction_id': tx['id']
            },
            'owners_before': output['public_keys']
        }
        prepared_transfer_tx = bdb_connection.transactions.prepare(
            operation='TRANSFER',
            asset=transfer_asset,
            metadata=metadata,
            inputs=transfer_input,
            recipients=recipient_pub_key
        )
        fulfilled_transfer_tx = bdb_connection.transactions.fulfill(
            prepared_transfer_tx,
            private_keys=self.private_key,
        )
        sent_transfer_tx = bdb_connection.transactions.send(fulfilled_transfer_tx)
        if fulfilled_transfer_tx != sent_transfer_tx:
            raise exceptions.CreateRecordException()
        return sent_transfer_tx

    def get_id(self) -> str:
        """
        Get Account id

        :return: Internal id of Account
        :rtype: str
        """
        return self.account_id

    def get_public_key(self) -> str:
        """
        Get public key

        :return: Public key of Account
        :rtype: str
        """
        return self.public_key


class DocumentConceptAccount(BaseAccount):
    """
    BigchainDB Document Concept Account
    """

    def __init__(self, account_id: str, store: local_stores.SqliteStore):
        """
        Instantiate Document Concept Account object

        :param account_id: Internal id of Account
        :type account_id: str
        :param store: Local database object
        :type store: local_stores.SqliteStore
        """
        super().__init__(account_id, store)

    def save_asset(self, asset: dict, bdb_connection: BigchainDB) -> str:
        """
        Write asset to BigchainDB

        :param asset: Dictonary with asset data
        :type asset: dict
        :param bdb_connection: Connection object for BigchainDB
        :type bdb_connection: BigchainDB
        :return: Transaction Id
        :rtype: str
        """
        asset = {'data': asset}
        metadata = {'account_id': self.account_id}
        tx = self._create_asset(bdb_connection, asset, metadata)
        utils.wait_until_valid(tx['id'], bdb_connection)
        tx = self._transfer_asset(bdb_connection, self.public_key, tx, metadata)
        log.info("Created document: %s - %s", self.account_id, tx['id'])
        return tx['id']


class GraphConceptAccount(BaseAccount):
    """
    BigchainDB Graph Concept Account
    """

    def __init__(self, prov_element: ProvElement, prov_relations: dict, id_mapping: dict, namespaces: list,
                 store: local_stores.SqliteStore = local_stores.SqliteStore()):
        """
        Instantiate Graph Concept Account object

        :param prov_element: ProvElement related to account
        :type prov_element: ProvElement
        :param prov_relations: List including dictionaries of all outgoing ProvRelations
        :type prov_relations: list
        :param namespaces: List of Prov Namespaces
        :type namespaces: list
        :param store: Local database object
        :type store: local_stores.SqliteStore
        """
        assert prov_element is not None
        assert prov_relations is not None
        assert namespaces is not None
        self.prov_element = prov_element
        self.prov_namespaces = namespaces
        self.prov_relations_with_id = prov_relations['with_id']
        self.id_mapping = id_mapping
        self.prov_relations_without_id = prov_relations['without_id']
        super().__init__(str(prov_element.identifier), store)

    def get_tx_id(self) -> str:
        """
        Get the tx_id that describes the account in BigchainDB

        :return: Transaction id of account
        :rtype: str
        """
        return self.tx_id

    def has_relations_with_id(self) -> bool:
        """
        Indicates whether an account has relation with ids or not
        :return: True if one or more relation does have ids
        :rtype: bool
        """
        return len(self.prov_relations_with_id) != 0

    def has_relations_without_id(self) -> bool:
        """
        Indicates whether an account has relation without ids or not
        :return: True if one or more relation does have ids
        :rtype: bool
        """
        return len(self.prov_relations_without_id) != 0

    def __str__(self):
        return "{} : {}\n\t{}\n\t{}".format(self.account_id, self.public_key, self.prov_relations_with_id,
                                            self.prov_relations_without_id)

    def __create_instance_document(self) -> ProvDocument:
        """
        Creates a valid ProvDocument describing an account

        :return: Representation of account as ProvDocument
        :rtype: ProvDocument
        """
        doc = ProvDocument()
        for n in self.prov_namespaces:
            doc.add_namespace(n.prefix, n.uri)
        doc.add_record(self.prov_element)
        return doc

    def __create_relation(self, relation) -> (ProvDocument, dict):
        """
        Returns a ProvDocument and mapping for all relations
        
        :return: Relation as ProvDocument and 
        :rtype: ( ProvDocument, dict)
        """
        doc = ProvDocument()
        mapping = {}
        for relation_type, relation_attr in relation.formal_attributes:
            if relation_attr:
                try:
                    recipient = self.store.get_account(str(relation_attr))
                    mapping[recipient[0]] = recipient[3]
                except exceptions.NoAccountFoundException:
                    try:
                        recipient = self.id_mapping.get(str(relation_attr))
                        mapping[str(relation_attr)] = recipient
                    except exceptions.NoRelationFoundException:
                        log.info("Found no tx for %s", relation_attr)

        for n in self.prov_namespaces:
            doc.add_namespace(n.prefix, n.uri)
        doc.add_record(relation)
        return doc, mapping

    def save_relations_with_ids(self, bdb_connection: BigchainDB) -> list:
        """
        Writes all assets with relations (having ids) to BigchainDB

        :param bdb_connection: Connection object for BigchainDB
        :type bdb_connection: BigchainDB
        :return: Transactions ids of all relations
        :rtype: list
        """
        if self.tx_id == '':
            raise exceptions.AccountNotCreatedException("Account must be created before transactions")
        tx_list = []
        for relation in self.prov_relations_with_id:
            doc, mapping = self.__create_relation(relation)
            for record in doc.get_records():
                recipient = self.store.get_account(str(record.args[1]))
                asset = {'data': {'prov': doc.serialize(format='json'), 'map': mapping}}
                metadata = {'relation': '->'.join([self.account_id, recipient[0]])}
                tx = self._create_asset(bdb_connection, asset, metadata)
                utils.wait_until_valid(tx['id'], bdb_connection)
                metadata = {'relation': '->'.join([self.account_id, recipient[0]])}
                tx = self._transfer_asset(bdb_connection, recipient[1], tx, metadata)
                tx_list.append(tx['id'])
                self.id_mapping[str(record.identifier)] = tx['id']
                log.debug("Created relation %s: %s -> %s - %s", record.identifier, self.account_id, recipient[0],
                          tx['id'])
        return tx_list

    def save_relations_without_ids(self, bdb_connection: BigchainDB) -> list:
        """
        Write all assets with relations to BigchainDB

        :param bdb_connection: Connection object for BigchainDB
        :type bdb_connection: BigchainDB
        :return: Transactions ids of all relations
        :rtype: list
        """
        if self.tx_id == '':
            raise exceptions.AccountNotCreatedException("Account must be created before transactions")
        tx_list = []
        for relation in self.prov_relations_without_id:
            doc, mapping = self.__create_relation(relation)
            for records in doc.get_records():
                recipient = self.store.get_account(str(records.args[1]))
                asset = {'data': {'prov': doc.serialize(format='json'), 'map': mapping}}
                metadata = {'relation': '->'.join([self.account_id, recipient[0]])}
                tx = self._create_asset(bdb_connection, asset, metadata)
                utils.wait_until_valid(tx['id'], bdb_connection)
                metadata = {'relation': '->'.join([self.account_id, recipient[0]])}
                tx = self._transfer_asset(bdb_connection, recipient[1], tx, metadata)
                tx_list.append(tx['id'])
                log.debug("Created relation: %s -> %s - %s", self.account_id, recipient[0], tx['id'])
        return tx_list

    def save_instance_asset(self, bdb_connection: BigchainDB) -> str:
        """
        Write provenance describing the account to BigchainDB

        :param bdb_connection: Connection object for BigchainDB
        :type bdb_connection: BigchainDB
        :return: Transactions id of instance
        :rtype: str
        """
        if self.tx_id == '':
            prov_document = self.__create_instance_document()
            asset = {'data': {'prov': prov_document.serialize(format='json')}}
            metadata = {'instance': self.account_id}
            tx = self._create_asset(bdb_connection, asset, metadata)
            utils.wait_until_valid(tx['id'], bdb_connection)
            tx = self._transfer_asset(bdb_connection, self.public_key, tx, metadata)
            self.store.write_tx_id(self.account_id, tx['id'])
            self.tx_id = tx['id']
            log.debug("Created instance: %s - %s", self.account_id, tx['id'])
        return self.tx_id


class RoleConceptAccount(BaseAccount):
    """
    BigchainDB Graph Concept Account
    """

    def __init__(self, agent: ProvAgent, relations: list, elements: dict, id_mapping: dict, namespaces: list,
                 store: local_stores.SqliteStore = local_stores.SqliteStore()):
        """
        Instantiate Graph Concept Account object

        :param agent: ProvAgent related to account
        :type agent: ProvAgent
        :param namespaces: List of Prov Namespaces
        :type namespaces: list
        :param store: Local database object
        :type store: local_stores.SqliteStore
        """
        assert agent is not None
        assert elements is not None
        assert namespaces is not None
        self.prov_agent = agent
        self.prov_agent_relations = relations
        self.prov_namespaces = namespaces
        self.prov_elements = elements
        self.id_mapping = id_mapping

        super().__init__(str(agent.identifier), store)

    def get_tx_id(self) -> str:
        """
        Get the tx_id that describes the account in BigchainDB

        :return: Transaction id of account
        :rtype: str
        """
        return self.tx_id

    def __str__(self):
        return "{} : {}\n\t{}\n\t{}".format(self.account_id, self.public_key,
                                            self.prov_agent_relations, self.prov_elements)

    def __create_document(self, element, relations) -> tuple:
        """
        Returns a ProvDocument and mapping for all relations

        :return: Relation as ProvDocument and
        :rtype: (ProvDocument, map)
        """
        doc = ProvDocument()
        doc.add_record(element)
        mapping = {}
        for relation in relations:
            for n in self.prov_namespaces:
                doc.add_namespace(n.prefix, n.uri)
            doc.add_record(relation)
        return doc, mapping

    def save_elements(self, bdb_connection: BigchainDB) -> list:
        """
        Writes all elements with assets to BigchainDB

        :param bdb_connection: Connection object for BigchainDB
        :type bdb_connection: BigchainDB
        :return: Transactions ids of all relations
        :rtype: list
        """
        if self.tx_id == '':
            raise exceptions.AccountNotCreatedException("Account must be created before transactions")
        tx_list = []
        for ordered_element in self.prov_elements.values():
            for element, relations in ordered_element.items():
                doc, mapping = self.__create_document(element, relations)
                asset = {'data': {'prov': doc.serialize(format='json'), 'map': mapping}}
                metadata = {'instance': self.account_id}
                tx = self._create_asset(bdb_connection, asset, metadata)
                utils.wait_until_valid(tx['id'], bdb_connection)
                tx = self._transfer_asset(bdb_connection, self.public_key, tx, metadata)
                self.store.write_account(str(element.identifier), '', '', tx['id'])
                tx_list.append(tx['id'])
                # for id, tx_id in mapping.items():
                #    if not tx_id:
                #        self.id_mapping[id] = tx['id']
                # self.id_mapping[str(element.identifier)] = tx['id']
                log.debug("Created element %s related to %s - %s", element.identifier, self.account_id, tx['id'])
        return tx_list

    def save_instance_asset(self, bdb_connection: BigchainDB) -> str:
        """
        Write provenance describing the account to BigchainDB

        :param bdb_connection: Connection object for BigchainDB
        :type bdb_connection: BigchainDB
        :return: Transactions id of instance
        :rtype: str
        """
        if self.tx_id == '':
            prov_document, mapping = self.__create_document(self.prov_agent, self.prov_agent_relations)
            asset = {'data': {'prov': prov_document.serialize(format='json'), 'map': mapping}}
            metadata = {'instance': self.account_id}
            tx = self._create_asset(bdb_connection, asset, metadata)
            utils.wait_until_valid(tx['id'], bdb_connection)
            tx = self._transfer_asset(bdb_connection, self.public_key, tx, metadata)
            self.store.write_tx_id(self.account_id, tx['id'])
            self.id_mapping[self.account_id] = tx['id']
            self.tx_id = tx['id']
            log.debug("Created agent: %s - %s", self.account_id, tx['id'])
        return self.tx_id



import logging
import unittest
from unittest import mock

from . import setup_test_files
import bigchaindb_driver
from prov2bigchaindb.core import clients, accounts, utils, exceptions, local_stores

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class BaseAccountTest(unittest.TestCase):
    def setUp(self):
        self.account_id = 'Base_Account_Test'
        self.public_key = 'public'
        self.private_key = 'private'

        self.store = mock.Mock(spec=local_stores.SqliteStore,
                               **{'get_account.return_value': (self.account_id,
                                                               self.public_key,
                                                               self.private_key,
                                                               None)}
                               )

    def tearDown(self):
        del self.account_id
        del self.public_key
        del self.private_key
        del self.store

    def test_positive_init(self):
        account = accounts.BaseAccount(self.account_id, self.store)
        self.assertIsInstance(account, accounts.BaseAccount)
        self.assertIsInstance(account.public_key, str)
        self.assertIsInstance(account.private_key, str)
        self.assertEqual(account.get_public_key(), self.public_key)

    def test_positive_init_without_Account(self):
        self.store.configure_mock(**{'get_account.side_effect': exceptions.NoAccountFoundException()})
        account = accounts.BaseAccount(self.account_id, self.store)
        self.assertNotEqual(account.get_public_key(), self.public_key)

    def test_negative_init_BaseAccount(self):
        with self.assertRaises(AssertionError):
            accounts.BaseAccount(self.account_id, None)

    def test__str__(self):
        account = accounts.BaseAccount(self.account_id, self.store)
        self.assertEqual(str(account), self.account_id + " : " + self.public_key)

    def test_positive_get_id(self):
        account = accounts.BaseAccount(self.account_id, self.store)
        self.assertIsInstance(account.get_id(), str)
        self.assertEqual(account.get_id(), self.account_id)

    def test_positive_get_public_Key(self):
        account = accounts.BaseAccount(self.account_id, self.store)
        self.assertIsInstance(account.get_public_key(), str)
        self.assertEqual(account.get_public_key(), self.public_key)

    @unittest.skip("testing skipping")
    def test__create_asset(self):
        raise NotImplementedError()

    @unittest.skip("testing skipping")
    def test__transfer_asset(self):
        raise NotImplementedError()


class DocumentConceptAccountTest(unittest.TestCase):
    def setUp(self):
        self.account_id = 'Document_Concept_Account_Test'
        self.public_key = 'public'
        self.private_key = 'private'
        self.store = mock.Mock(spec=local_stores.SqliteStore,
                               **{'get_account.return_value': (
                                   self.account_id,
                                   self.public_key,
                                   self.private_key,
                                   None)})

        self.bdb_returned_transaction = {"operation": "CREATE",
                                         "outputs": [{"amount": 1,
                                                      "condition": {
                                                          "details": {
                                                              "bitmask": 32,
                                                              "public_key": "4K9sWUMFwTgaDGPfdynrbxWqWS6sWmKbZoTjxLtVUibD",
                                                              "signature": None,
                                                              "type": "fulfillment",
                                                              "type_id": 4
                                                          },
                                                          "uri": "cc:4:20:MTmLrdyfhfxPw3WxnaYaQkPmU1GcEzg9mAj_O_Nuv5w:96"
                                                      },
                                                      "public_keys": [
                                                          "4K9sWUMFwTgaDGPfdynrbxWqWS6sWmKbZoTjxLtVUibD"
                                                      ]
                                                      }
                                                     ],
                                         'id': '1',
                                         'asset': {'data': {'prov': '1'}}}

        self.bdb_connection = mock.Mock(spec=bigchaindb_driver.BigchainDB,
                                        **{'transactions.retrieve.return_value': self.bdb_returned_transaction,
                                           'transactions.prepare.return_value': self.bdb_returned_transaction,
                                           'transactions.fulfill.return_value': self.bdb_returned_transaction,
                                           'transactions.send.return_value': self.bdb_returned_transaction,
                                           }
                                        )

    def tearDown(self):
        del self.account_id
        del self.public_key
        del self.private_key
        del self.store
        del self.bdb_connection
        del self.bdb_returned_transaction

    @mock.patch('prov2bigchaindb.core.utils.wait_until_valid')
    def test_positive_save_asset(self, mock_wait):
        asset = {'data': {'prov': ''}}
        account = accounts.DocumentConceptAccount(self.account_id, self.store)
        tx_id = account.save_asset(asset, self.bdb_connection)
        # self.bdb_connection.transactions.prepare.assert_called_with(operation='CREATE', signers=self.public_key, asset={'data':asset}, metadata={'account_id':self.account_id})
        self.bdb_connection.transactions.fulfill.assert_called_with(self.bdb_returned_transaction,
                                                                    private_keys=self.private_key)
        self.bdb_connection.transactions.send.assert_called_with(self.bdb_returned_transaction)
        self.assertEqual(tx_id, '1')

    @mock.patch('prov2bigchaindb.core.utils.wait_until_valid')
    def test_negative_save_asset(self, mock_wait):
        asset = {'data': {'prov': ''}}
        account = accounts.DocumentConceptAccount(self.account_id, self.store)
        negative_return = self.bdb_returned_transaction['id'] = '2'
        self.bdb_connection.configure_mock(**{'transactions.retrieve.return_value': self.bdb_returned_transaction,
                                              'transactions.prepare.return_value': self.bdb_returned_transaction,
                                              'transactions.fulfill.return_value': self.bdb_returned_transaction,
                                              'transactions.send.return_value': negative_return,
                                              })
        with self.assertRaises(exceptions.CreateRecordException):
            account.save_asset(asset, self.bdb_connection)
        # self.bdb_connection.transactions.prepare.assert_called_with(operation='CREATE', signers=self.public_key, asset={'data':asset}, metadata={'account_id':self.account_id})
        self.bdb_connection.transactions.fulfill.assert_called_with(self.bdb_returned_transaction,
                                                                    private_keys=self.private_key)
        self.bdb_connection.transactions.send.assert_called_with(self.bdb_returned_transaction)


class GraphConceptAccountTest(unittest.TestCase):
    def setUp(self):
        self.test_prov_files = setup_test_files()
        self.prov_document = utils.to_prov_document(content=self.test_prov_files["simple"])
        self.prov_element, self.prov_relations, self.prov_namespaces = \
            clients.GraphConceptClient.calculate_account_data(self.prov_document)[
                0]
        self.id_mapping = {}
        for rel in self.prov_relations['with_id']:
            self.id_mapping[rel.identifier] = ''
        self.public_key = 'public'
        self.private_key = 'private'
        self.store = mock.Mock(spec=local_stores.SqliteStore,
                               **{'get_account.return_value': (
                                   str(self.prov_element.identifier),
                                   self.public_key,
                                   self.private_key,
                                   None)})

        self.bdb_returned_transaction = {"operation": "CREATE",
                                         "outputs": [{"amount": 1,
                                                      "condition": {
                                                          "details": {
                                                              "bitmask": 32,
                                                              "public_key": "4K9sWUMFwTgaDGPfdynrbxWqWS6sWmKbZoTjxLtVUibD",
                                                              "signature": None,
                                                              "type": "fulfillment",
                                                              "type_id": 4
                                                          },
                                                          "uri": "cc:4:20:MTmLrdyfhfxPw3WxnaYaQkPmU1GcEzg9mAj_O_Nuv5w:96"
                                                      },
                                                      "public_keys": [
                                                          "4K9sWUMFwTgaDGPfdynrbxWqWS6sWmKbZoTjxLtVUibD"
                                                      ]
                                                      }
                                                     ],
                                         'id': '1',
                                         'asset': {'data': {'prov': '1'}}
                                         }
        self.bdb_connection = mock.Mock(spec=bigchaindb_driver.BigchainDB,
                                        **{'transactions.retrieve.return_value': self.bdb_returned_transaction,
                                           'transactions.prepare.return_value': self.bdb_returned_transaction,
                                           'transactions.fulfill.return_value': self.bdb_returned_transaction,
                                           'transactions.send.return_value': self.bdb_returned_transaction})

    def tearDown(self):
        del self.prov_document
        del self.prov_element
        del self.prov_relations
        del self.prov_namespaces
        del self.public_key
        del self.private_key
        del self.store
        del self.bdb_connection
        del self.bdb_returned_transaction
        del self.test_prov_files

    def test_positive_init_without_account(self):
        self.store.configure_mock(**{'get_account.side_effect': exceptions.NoAccountFoundException()})
        account = accounts.GraphConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                               self.prov_namespaces, self.store)
        self.assertNotEqual(account.get_public_key(), self.public_key)
        self.assertEqual(account.tx_id, '')
        self.assertEqual(account.prov_namespaces, self.prov_namespaces)
        self.assertEqual(account.prov_relations_without_id, self.prov_relations['without_id'])
        self.assertEqual(account.prov_relations_with_id, self.prov_relations['with_id'])

    def test__str__(self):
        account = accounts.GraphConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                               self.prov_namespaces, self.store)
        self.assertEqual(str(account), str(self.prov_element.identifier) + " : " + self.public_key +
                         "\n\t" + str(self.prov_relations['with_id']) +
                         "\n\t" + str(self.prov_relations['without_id'])
                         )

    def test_get_tx_id(self):
        account = accounts.GraphConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                               self.prov_namespaces, self.store)
        account.tx_id = '1'
        self.assertEqual(account.get_tx_id(), '1')

    @mock.patch('prov2bigchaindb.core.utils.wait_until_valid')
    def test_positiv_save_instance_asset(self, mock_wait):
        self.store.configure_mock(**{'get_account.side_effect': exceptions.NoAccountFoundException()})
        account = accounts.GraphConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                               self.prov_namespaces, self.store)
        tx_id = account.save_instance_asset(self.bdb_connection)
        # asset = {'data': {'prov': account._create_instance_document().serialize(format='json')}}
        # self.bdb_connection.transactions.prepare.assert_called_with(operation='CREATE', signers=account.public_key, asset=asset, metadata={'account_id':str(self.prov_element.identifier)})
        self.bdb_connection.transactions.fulfill.assert_called_with(self.bdb_returned_transaction,
                                                                    private_keys=account.private_key)
        self.bdb_connection.transactions.send.assert_called_with(self.bdb_returned_transaction)
        self.assertEqual(tx_id, '1')
        self.assertEqual(account.get_tx_id(), '1')

    @mock.patch('prov2bigchaindb.core.utils.wait_until_valid')
    def test_negativ_save_instance_asset(self, mock_wait):
        self.store.configure_mock(**{'get_account.side_effect': exceptions.NoAccountFoundException()})
        account = accounts.GraphConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                               self.prov_namespaces, self.store)
        negative_return = self.bdb_returned_transaction['id'] = '2'
        self.bdb_connection.configure_mock(**{'transactions.retrieve.return_value': self.bdb_returned_transaction,
                                              'transactions.prepare.return_value': self.bdb_returned_transaction,
                                              'transactions.fulfill.return_value': self.bdb_returned_transaction,
                                              'transactions.send.return_value': negative_return,
                                              })
        # asset = {'data': {'prov': account._create_instance_document().serialize(format='json')}}
        with self.assertRaises(exceptions.CreateRecordException):
            account.save_instance_asset(self.bdb_connection)
        # self.bdb_connection.transactions.prepare.assert_called_with(operation='CREATE', signers=account.public_key, asset=asset, metadata={'account_id':str(self.prov_element.identifier)})
        self.bdb_connection.transactions.fulfill.assert_called_with(self.bdb_returned_transaction,
                                                                    private_keys=account.private_key)
        self.bdb_connection.transactions.send.assert_called_with(self.bdb_returned_transaction)
        self.assertEqual(account.tx_id, '')

    @unittest.skip("testing skipping")
    def test_save_relations_assets(self):
        raise NotImplementedError()

    @unittest.skip("testing skipping")
    def test__create_instance_document(self):
        raise NotImplementedError()

    @unittest.skip("testing skipping")
    def test__create_relations_document(self):
        raise NotImplementedError()


class RoleConceptAccountTest(unittest.TestCase):
    def setUp(self):
        self.test_prov_files = setup_test_files()
        self.prov_document = utils.to_prov_document(content=self.test_prov_files["simple"])
        self.prov_element, self.prov_relations, self.prov_namespaces = \
            clients.RoleConceptClient.calculate_account_data(self.prov_document)[
                0]
        self.id_mapping = {}
        for rel in self.prov_relations['with_id']:
            self.id_mapping[rel.identifier] = ''
        self.public_key = 'public'
        self.private_key = 'private'
        self.store = mock.Mock(spec=local_stores.SqliteStore,
                               **{'get_account.return_value': (
                                   str(self.prov_element.identifier),
                                   self.public_key,
                                   self.private_key,
                                   None)})

        self.bdb_returned_transaction = {"operation": "CREATE",
                                         "outputs": [{"amount": 1,
                                                      "condition": {
                                                          "details": {
                                                              "bitmask": 32,
                                                              "public_key": "4K9sWUMFwTgaDGPfdynrbxWqWS6sWmKbZoTjxLtVUibD",
                                                              "signature": None,
                                                              "type": "fulfillment",
                                                              "type_id": 4
                                                          },
                                                          "uri": "cc:4:20:MTmLrdyfhfxPw3WxnaYaQkPmU1GcEzg9mAj_O_Nuv5w:96"
                                                      },
                                                      "public_keys": [
                                                          "4K9sWUMFwTgaDGPfdynrbxWqWS6sWmKbZoTjxLtVUibD"
                                                      ]
                                                      }
                                                     ],
                                         'id': '1',
                                         'asset': {'data': {'prov': '1'}}
                                         }
        self.bdb_connection = mock.Mock(spec=bigchaindb_driver.BigchainDB,
                                        **{'transactions.retrieve.return_value': self.bdb_returned_transaction,
                                           'transactions.prepare.return_value': self.bdb_returned_transaction,
                                           'transactions.fulfill.return_value': self.bdb_returned_transaction,
                                           'transactions.send.return_value': self.bdb_returned_transaction})

    def tearDown(self):
        del self.prov_document
        del self.prov_element
        del self.prov_relations
        del self.prov_namespaces
        del self.public_key
        del self.private_key
        del self.store
        del self.bdb_connection
        del self.bdb_returned_transaction
        del self.test_prov_files

    @unittest.skip("testing skipping")
    def test_positive_init_without_account(self):
        self.store.configure_mock(**{'get_account.side_effect': exceptions.NoAccountFoundException()})
        account = accounts.RoleConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                              self.prov_namespaces, self.store)
        self.assertNotEqual(account.get_public_key(), self.public_key)
        self.assertEqual(account.tx_id, '')
        self.assertEqual(account.prov_namespaces, self.prov_namespaces)
        self.assertEqual(account.prov_relations_without_id, self.prov_relations['without_id'])
        self.assertEqual(account.prov_relations_with_id, self.prov_relations['with_id'])

    @unittest.skip("testing skipping")
    def test__str__(self):
        account = accounts.RoleConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                              self.prov_namespaces, self.store)
        self.assertEqual(str(account), str(self.prov_element.identifier) + " : " + self.public_key +
                         "\n\t" + str(self.prov_relations['with_id']) +
                         "\n\t" + str(self.prov_relations['without_id'])
                         )

    @unittest.skip("testing skipping")
    def test_get_tx_id(self):
        account = accounts.RoleConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                              self.prov_namespaces, self.store)
        account.tx_id = '1'
        self.assertEqual(account.get_tx_id(), '1')

    @unittest.skip("testing skipping")
    @mock.patch('prov2bigchaindb.core.utils.wait_until_valid')
    def test_positiv_save_instance_asset(self, mock_wait):
        self.store.configure_mock(**{'get_account.side_effect': exceptions.NoAccountFoundException()})
        account = accounts.RoleConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                              self.prov_namespaces, self.store)
        tx_id = account.save_instance_asset(self.bdb_connection)
        # asset = {'data': {'prov': account._create_instance_document().serialize(format='json')}}
        # self.bdb_connection.transactions.prepare.assert_called_with(operation='CREATE', signers=account.public_key, asset=asset, metadata={'account_id':str(self.prov_element.identifier)})
        self.bdb_connection.transactions.fulfill.assert_called_with(self.bdb_returned_transaction,
                                                                    private_keys=account.private_key)
        self.bdb_connection.transactions.send.assert_called_with(self.bdb_returned_transaction)
        self.assertEqual(tx_id, '1')
        self.assertEqual(account.get_tx_id(), '1')

    @unittest.skip("testing skipping")
    @mock.patch('prov2bigchaindb.core.utils.wait_until_valid')
    def test_negativ_save_instance_asset(self, mock_wait):
        self.store.configure_mock(**{'get_account.side_effect': exceptions.NoAccountFoundException()})
        account = accounts.RoleConceptAccount(self.prov_element, self.prov_relations, self.id_mapping,
                                              self.prov_namespaces, self.store)
        negative_return = self.bdb_returned_transaction['id'] = '2'
        self.bdb_connection.configure_mock(**{'transactions.retrieve.return_value': self.bdb_returned_transaction,
                                              'transactions.prepare.return_value': self.bdb_returned_transaction,
                                              'transactions.fulfill.return_value': self.bdb_returned_transaction,
                                              'transactions.send.return_value': negative_return,
                                              })
        # asset = {'data': {'prov': account._create_instance_document().serialize(format='json')}}
        with self.assertRaises(exceptions.CreateRecordException):
            account.save_instance_asset(self.bdb_connection)
        # self.bdb_connection.transactions.prepare.assert_called_with(operation='CREATE', signers=account.public_key, asset=asset, metadata={'account_id':str(self.prov_element.identifier)})
        self.bdb_connection.transactions.fulfill.assert_called_with(self.bdb_returned_transaction,
                                                                    private_keys=account.private_key)
        self.bdb_connection.transactions.send.assert_called_with(self.bdb_returned_transaction)
        self.assertEqual(account.tx_id, '')

    @unittest.skip("testing skipping")
    def test_save_relations_assets(self):
        raise NotImplementedError()

    @unittest.skip("testing skipping")
    def test__create_instance_document(self):
        raise NotImplementedError()

    @unittest.skip("testing skipping")
    def test__create_relations_document(self):
        raise NotImplementedError()



import logging
from io import BufferedReader

import bigchaindb_driver as bd
import prov.graph as provgraph
import prov.model as provmodel
from bigchaindb_driver import pool as bdpool
from networkx import is_directed_acyclic_graph
from networkx import isolates
from networkx import topological_sort

from prov2bigchaindb.core import utils, local_stores, accounts

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class BaseClient(object):
    """ BigchainDB Base Client """

    def __init__(self, host: str = '0.0.0.0', port: int = 9984,
                 num_connections: int = 5, local_store: local_stores.SqliteStore = local_stores.SqliteStore()):
        """
        Instantiate Base Client object

        :param host: BigchaindDB Hostname or IP (default: 0.0.0.0)
        :type host: str
        :param port: BigchaindDB Port (default: 9984)
        :type port: int
        :param num_connections: Amount of connections made to BigchainDB node
        :type num_connections: int
        :param local_store: Local database object
        :type local_store: SqliteStore
        """
        assert num_connections > 0
        self.node = 'http://{}:{}'.format(host, str(port))
        self.connections = num_connections * [bd.BigchainDB(self.node)]
        self.connection_pool = bdpool.Pool(self.connections)
        self.store = local_store

    def test_transaction(self, tx: dict) -> bool:
        """
        Validate a transaction against BigchainDB

        :param tx: Transaction to test
        :type tx: dict
        :return: True or Exception
        :rtype: bool
        """
        reason = None
        if not utils.is_valid_tx(tx['id'], self.connection_pool.get_connection()):
            reason = "TX is invalid"
        elif not utils.is_block_to_tx_valid(tx['id'], self.connection_pool.get_connection()):
            reason = "Block is invalid"
        if reason is None:
            return True
        log.error("Test failed: %s", tx['id'])
        raise Exception(reason)

    def _get_bigchain_connection(self) -> bd.BigchainDB:
        """
        Returns BigchainDB connection
        :return: BigchainDB connection object
        :rtype: bd.BigchainDB
        """
        return self.connection_pool.get_connection()

    def save_document(self, document: object) -> object:
        """
        Abstract method to store a document

        :param document: Document to save
        :type document: object
        :return: id
        :rtype: object
        """
        raise NotImplementedError("Abstract method")

    def get_document(self, document_id: object) -> provmodel.ProvDocument:
        """
        Abstract method to retrieve a document

        :param document_id: Document to save
        :type document_id: object
        :rtype: ProvDocument
        """
        raise NotImplementedError("Abstract method")


class DocumentConceptClient(BaseClient):
    """"""

    def __init__(self, account_id: str = None, host: str = '0.0.0.0', port: int = 9984, num_connections: int = 1,
                 local_store: local_stores.SqliteStore = local_stores.SqliteStore()):
        """
        Instantiate Document Client object

        :param host: BigchaindDB Hostname or IP (default: 0.0.0.0)
        :type host: str
        :param port: BigchaindDB Port (default: 9984)
        :type port: int
        :param local_store: Local database object
        :type local_store: SqliteStore
        """
        super().__init__(host, port, num_connections, local_store)
        self.account = accounts.DocumentConceptAccount(account_id, self.store)

    def save_document(self, document: str or bytes or provmodel.ProvDocument) -> str:
        """
        Write a document into BigchainDB
        
        :param document: Document as JSON/XML/PROVN
        :type document: str or bytes or ProvDocument
        :return: Transaction id of document
        :rtype: str
        """
        log.info("Save document...")
        prov_document = utils.to_prov_document(content=document)
        asset = {'prov': prov_document.serialize(format='json')}
        tx_id = self.account.save_asset(asset, self._get_bigchain_connection())
        log.info("Saved document in Tx with id: %s", tx_id)
        return tx_id

    def get_document(self, tx_id: str) -> provmodel.ProvDocument:
        """
        Retrieve a document by transaction id from BigchainDB

        :param tx_id: Transaction Id of Document
        :type tx_id: str
        :return: Document as ProvDocument object
        :rtype: ProvDocument
        """
        log.info("Retrieve and build document")
        tx = self._get_bigchain_connection().transactions.retrieve(tx_id)
        self.test_transaction(tx)
        if 'id' in tx['asset'].keys():
            tx = self._get_bigchain_connection().transactions.get(asset_id=tx['asset']['id'])[0]
            self.test_transaction(tx)
        log.info("Success")
        return utils.to_prov_document(tx['asset']['data']['prov'])


class GraphConceptClient(BaseClient):
    """"""

    def __init__(self, host: str = '0.0.0.0', port: int = 9984, num_connections: int = 5,
                 local_store: local_stores.SqliteStore = local_stores.SqliteStore()):
        """
        Instantiate Graph Client object

        :param host: BigchaindDB Hostname or IP (default: 0.0.0.0)
        :type host: str
        :param port: BigchaindDB Port (default: 9984)
        :type port: int
        :param local_store: Local database object
        :type local_store: SqliteStore
        """
        super().__init__(host, port, num_connections, local_store=local_store)
        self.accounts = []

    @staticmethod
    def calculate_account_data(prov_document: provmodel.ProvDocument) -> list:
        """
        Transforms a ProvDocument into a tuple with ProvElement, list of ProvRelation and list of Namespaces

        :param prov_document: Document to transform
        :type prov_document:
        :return: List of tuples(element, relations, namespace)
        :rtype: list
        """

        namespaces = prov_document.get_registered_namespaces()
        g = provgraph.prov_to_graph(prov_document=prov_document)
        elements = []
        for node, nodes in g.adjacency_iter():
            relations = {'with_id': [], 'without_id': []}
            # print(node)
            for tmp_relations in nodes.values():
                for relation in tmp_relations.values():
                    relation = relation['relation']
                    if relation.identifier:
                        relations['with_id'].append(relation)
                    else:
                        relations['without_id'].append(relation)
            elements.append((node, relations, namespaces))
        return elements

    def save_document(self, document: str or BufferedReader or provmodel.ProvDocument) -> list:
        """
        Write a document into BigchainDB

        :param document: Document as JSON/XML/PROVN
        :type document: str or BufferedReader or ProvDocument
        :return: List of transaction ids
        :rtype: list
        """
        log.info("Save document...")
        document_tx_ids = []
        prov_document = utils.to_prov_document(content=document)
        elements = GraphConceptClient.calculate_account_data(prov_document)
        id_mapping = {}
        log.info("Create and Save instances")
        for prov_element, prov_relations, namespaces in elements:
            for rel in prov_relations['with_id']:
                id_mapping[rel.identifier] = ''

        for prov_element, prov_relations, namespaces in elements:
            account = accounts.GraphConceptAccount(prov_element, prov_relations, id_mapping, namespaces, self.store)
            self.accounts.append(account)
            tx_id = account.save_instance_asset(self._get_bigchain_connection())
            document_tx_ids.append(tx_id)

        log.info("Save relations with ids")
        for account in filter(lambda acc: acc.has_relations_with_id, self.accounts):
            document_tx_ids += account.save_relations_with_ids(self._get_bigchain_connection())

        log.info("Save relations without ids")
        for account in filter(lambda acc: acc.has_relations_without_id, self.accounts):
            document_tx_ids += account.save_relations_without_ids(self._get_bigchain_connection())

        log.info("Saved document in %s Tx", len(document_tx_ids))
        return document_tx_ids

    def get_document(self, document_tx_ids: list) -> provmodel.ProvDocument:
        """
        Retrieve a document by a list transaction ids from BigchainDB

        :param document_tx_ids: Transaction Ids of Document
        :type document_tx_ids: list
        :return: Document as ProvDocument object
        :rtype: ProvDocument
        """
        log.info("Retrieve and rebuild document...")
        doc = provmodel.ProvDocument()
        for i in document_tx_ids:
            log.info("tx id: %s",i)
            tx = self._get_bigchain_connection().transactions.get(asset_id=i)[0]
            self.test_transaction(tx)
            if 'id' in tx['asset'].keys():
                tx = self._get_bigchain_connection().transactions.get(asset_id=tx['asset']['id'])[0]
                self.test_transaction(tx)
            tmp_doc = utils.to_prov_document(tx['asset']['data']['prov'])
            for namespace in tmp_doc.get_registered_namespaces():
                doc.add_namespace(namespace)
            for record in tmp_doc.get_records():
                doc.add_record(record=record)
        log.info("Success")
        return doc


class RoleConceptClient(BaseClient):
    """"""

    def __init__(self, host: str = '0.0.0.0', port: int = 9984, num_connections: int = 5,
                 local_store: local_stores.SqliteStore = local_stores.SqliteStore()):
        """
        Instantiate Role Client object

        :param host: BigchaindDB Hostname or IP (default: 0.0.0.0)
        :type host: str
        :param port: BigchaindDB Port (default: 9984)
        :type port: int
        :param local_store: Local database object
        :type local_store: SqliteStore
        """
        super().__init__(host, port, num_connections, local_store=local_store)
        self.accounts = []

    @staticmethod
    def calculate_account_data(prov_document: provmodel.ProvDocument) -> list:
        """
        Transforms a ProvDocument into a list of tuples including: 
        ProvAgent, list of ProvRelations from agent,
        list of ProvElements associated to ProvAgent,
        list of Namespaces

        :param prov_document: Document to transform
        :type prov_document:
        :return: List of tuples(ProvAgent, list(), list(), list())
        :rtype: list
        """

        namespaces = prov_document.get_registered_namespaces()
        g = provgraph.prov_to_graph(prov_document=prov_document)
        sorted_nodes = topological_sort(g, reverse=True)
        agents = list(filter(lambda elem: isinstance(elem, provmodel.ProvAgent), sorted_nodes))
        elements = list(filter(lambda elem: not isinstance(elem, provmodel.ProvAgent), sorted_nodes))

        # Check on compatibility
        if not is_directed_acyclic_graph(g):
            raise Exception("Provenance graph is not acyclic")
        if isolates(g):
            raise Exception("Provenance not compatible with role-based concept. Has isolated Elements")
        for element in elements:
            if provmodel.ProvAgent not in [type(n) for n in g.neighbors(element)]:
                raise Exception(
                    "Provenance not compatible with role-based concept. Element {} has not relation to any agent".format(
                        element))

        accounts = []
        for agent in agents:
            # find out-going relations from agent
            agent_relations = []
            for u, v in g.out_edges(agent):
                # Todo check if filter does not left out some info
                agent_relations.append(g.get_edge_data(u, v)[0]['relation'])

            agent_elements = {}
            i = 0
            for element in elements:
                element_relations = []
                if g.has_edge(element, agent):
                    for u, v in set(g.out_edges(element)):
                        for relation in g[u][v].values():
                            element_relations.append(relation['relation'])
                    agent_elements[i] = {element: element_relations}
                    i += 1

            accounts.append((agent, agent_relations, agent_elements, namespaces))
        return accounts

    def save_document(self, document: str or BufferedReader or provmodel.ProvDocument) -> list:
        """
        Write a document into BigchainDB

        :param document: Document as JSON/XML/PROVN
        :type document: str or BufferedReader or ProvDocument
        :return: List of transaction ids
        :rtype: list
        """
        log.info("Save document...")
        document_tx_ids = []
        prov_document = utils.to_prov_document(content=document)
        account_data = RoleConceptClient.calculate_account_data(prov_document)

        id_mapping = {}
        log.info("Create and Save instances")
        for agent, relations, elements, namespaces in account_data:
            account = accounts.RoleConceptAccount(agent, relations, elements, id_mapping, namespaces, self.store)
            self.accounts.append(account)
            tx_id = account.save_instance_asset(self._get_bigchain_connection())
            document_tx_ids.append(tx_id)

        log.info("Save elements")
        for account in self.accounts:
            document_tx_ids += account.save_elements(self._get_bigchain_connection())

        log.info("Saved document in %s Tx", len(document_tx_ids))
        return document_tx_ids

    def get_document(self, document_tx_ids: list) -> provmodel.ProvDocument:
        """
        Returns a document by a list transaction ids from BigchainDB

        :param document_tx_ids: Transaction Ids of Document
        :type document_tx_ids: list
        :return: Document as ProvDocument object
        :rtype: ProvDocument
        """
        log.info("Retrieve and rebuild document...")
        doc = provmodel.ProvDocument()
        for i in document_tx_ids:
            tx = self._get_bigchain_connection().transactions.get(asset_id=i)[0]
            self.test_transaction(tx)
            if 'id' in tx['asset'].keys():
                tx = self._get_bigchain_connection().transactions.get(asset_id=tx['asset']['id'])[0]
                self.test_transaction(tx)
            tmp_doc = utils.to_prov_document(tx['asset']['data']['prov'])
            for namespace in tmp_doc.get_registered_namespaces():
                doc.add_namespace(namespace)
            for record in tmp_doc.get_records():
                doc.add_record(record=record)
        log.info("Success")
        return doc


















class Prov2BigchainDBException(Exception):
    pass


class CreateRecordException(Prov2BigchainDBException):
    pass


class ParseException(Prov2BigchainDBException):
    pass


class NoAccountFoundException(Prov2BigchainDBException):
    pass


class AccountNotCreatedException(Prov2BigchainDBException):
    pass


class NoRelationFoundException(Prov2BigchainDBException):
    pass


class TransactionIdNotFound(Prov2BigchainDBException):
    pass


class BlockIdNotFound(Prov2BigchainDBException):
    pass



#!/usr/bin/env python
try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

install_requires=[
    "bigchaindb_driver~=0.3",
    "prov"
]

tests_require = [
    "BigchainDB~=1.0.0rc1",
    'coverage',
    'coveralls'
]

docs_require = [
    'Sphinx>=1.3.5',
    'recommonmark>=0.4.0',
    'sphinx-rtd-theme>=0.1.9',
    'sphinxcontrib-napoleon>=0.4.4',
    'sphinxcontrib-httpdomain>=1.5.0',
]


# get version
version = {}
with open('prov2bigchaindb/version.py') as fp:
    exec(fp.read(), version)


setup(
    name='prov2bigchaindb',
    version=version['__version__'],
    description='PROV 2 BigchainDB',
    long_description=__doc__,
    keywords=[
        'provenance', 'BigchainDB', 'blockchain', 'PROV', 'PROV-DM', 'PROV-JSON',
        'PROV-XML', 'PROV-N'
    ],
    author='Martin Stoffers',
    author_email='martin.stoffers@dlr.de',
    url='https://github.com/DLR-SC/prov2bigchaindb',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],

    packages=find_packages(),
    package_dir={
        'prov2bigchaindb': 'prov2bigchaindb'
    },
    include_package_data=True,
    zip_safe=False,

    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        'test': tests_require + docs_require,
        'dev': tests_require + docs_require,
        'docs': docs_require,
    },

    license="Apache License 2.0",
    test_suite='prov2bigchaindb.tests',
)



import pkg_resources


def setup_test_files():
    test_resources = {
        'simple': {'package': 'prov2bigchaindb', 'file': '/assets/graph-example.json'},
        'simple2': {'package': 'prov2bigchaindb', 'file': '/assets/test-example.json'},
        'quantified': {'package': 'prov2bigchaindb', 'file': '/assets/quantified-self.json'},
        'thesis': {'package': 'prov2bigchaindb', 'file': '/assets/thesis-example-full.json'}
    }
    return dict(
        (key, pkg_resources.resource_string(val['package'], val['file'])) for key, val in test_resources.items()
    )
